OPERATOR_DICT = {
    "Conv": {
        "Zero-Invariant": "TRUE",
        "Type": "Stem",
        "Shape-Dependent": "N/A"
    },
    "Gemm": {
        "Zero-Invariant": "TRUE",
        "Type": "Stem",
        "Shape-Dependent": "N/A"
    },
    "Linear": {
        "Zero-Invariant": "TRUE",
        "Type": "Stem",
        "Shape-Dependent": "N/A"
    },
    "BatchNormalization": {
        "Zero-Invariant": "TRUE",
        "Type": "Auxilary",
        "Shape-Dependent": "N/A"
    },
    "BatchNorm": {
        "Zero-Invariant": "TRUE",
        "Type": "Auxilary",
        "Shape-Dependent": "N/A"
    },
    "Relu": {
        "Zero-Invariant": "TRUE",
        "Type": "Activation",
        "Shape-Dependent": "N/A"
    },
    "LeakyRelu": {
        "Zero-Invariant": "TRUE",
        "Type": "Activation",
        "Shape-Dependent": "N/A"
    },
    "Abs": {
        "Zero-Invariant": "TRUE",
        "Type": "Activation",
        "Shape-Dependent": "N/A"
    },
    "Sin": {
        "Zero-Invariant": "TRUE",
        "Type": "Activation",
        "Shape-Dependent": "N/A"
    },
    "Sign": {
        "Zero-Invariant": "TRUE",
        "Type": "Activation",
        "Shape-Dependent": "N/A"
    },
    "Sqrt": {
        "Zero-Invariant": "TRUE",
        "Type": "Activation",
        "Shape-Dependent": "N/A"
    },
    "Pad": {
        "Zero-Invariant": "TRUE",
        "Type": "Auxiliary",
        "Shape-Dependent": "N/A"
    },
    "Add": {
        "Zero-Invariant": "TRUE",
        "Type": "Joint",
        "Shape-Dependent": "TRUE"
    },
    "Mul": {
        "Zero-Invariant": "TRUE",
        "Type": "Joint",
        "Shape-Dependent": "TRUE"
    },
    "Concat": {
        "Zero-Invariant": "TRUE",
        "Type": "Joint",
        "Shape-Dependent": "FALSE"
    },
    "MaxPool": {
        "Zero-Invariant": "TRUE",
        "Type": "Pool",
        "Shape-Dependent": "FALSE"
    },
    "AveragePool": {
        "Zero-Invariant": "TRUE",
        "Type": "Pool",
        "Shape-Dependent": "FALSE"
    },
    "Dropout": {
        "Zero-Invariant": "TRUE",
        "Type": "Dropout",
        "Shape-Dependent": "FALSE"
    },
    "Shape": {
        "Zero-Invariant": "TRUE",
        "Type": "Others",
        "Shape-Dependent": "FALSE"
    },
    "Reshape": {
        "Zero-Invariant": "TRUE",
        "Type": "Others",
        "Shape-Dependent": "FALSE"
    },
    "Transpose": {
        "Zero-Invariant": "FALSE",
        "Type": "Others",
        "Shape-Dependent": "FALSE"
    },
    "Constant": {
        "Zero-Invariant": "TRUE",
        "Type": "Others",
        "Shape-Dependent": ""
    },   
    "Upsample": {
        "Zero-Invariant": "TRUE",
        "Type": "Others",
        "Shape-Dependent": ""
    }, 
    "Gather": {
        "Zero-Invariant": "TRUE",
        "Type": "Others",
        "Shape-Dependent": "",
        "NOT-SURE": "TRUE"
    },   
    "Unsqueeze": {
        "Zero-Invariant": "TRUE",
        "Type": "Others",
        "Shape-Dependent": "",
        "NOT-SURE": "TRUE"
    },   
    "GlobalAveragePool": {
        "Zero-Invariant": "TRUE",
        "Type": "Pool",
        "Shape-Dependent": "",
        "NOT-SURE": "TRUE"
    },   
    "": {
        "Zero-Invariant": "",
        "Type": "",
        "Shape-Dependent": ""
    }

}


DEFAULT_OPT_PARAMS = {
    "sgd": {
        "first_momentum": 0.0,
        "second_momentum": 0.0,
        "dampening": 0.0,
        "weight_decay": 0.0,
        "lmbda": 1e-3,
        "lmbda_amplify": 2,
        "hat_lmbda_coeff": 10
    }
    ,
    "adam": {
        "lr": 1e-3,
        "first_momentum": 0.9,
        "second_momentum": 0.999,
        "dampening": 0.0,
        "weight_decay": 0.0,
        "lmbda": 1e-2,
        "lmbda_amplify": 20,
        "hat_lmbda_coeff": 1e3
    }
    ,
    "adamw": {
        "lr": 1e-3,
        "first_momentum": 0.9,
        "second_momentum": 0.999,
        "dampening": 0.0,
        "weight_decay": 1e-2,
        "lmbda": 1e-2,
        "lmbda_amplify": 20,
        "hat_lmbda_coeff": 1e3
    }
}



import os
import sys
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

from graph.connected_component import ConnectedComponent

def get_non_stem_nodes(graph):
    non_stem_nodes = list()
    for node in graph.nodes.values():
        if not node.is_stem() and not node.is_concat(axis=1): # operator concat is left for next version
            non_stem_nodes.append(node)
    return non_stem_nodes

def get_connected_components(graph, nodes):
    connected_components = []
    visited = dict()
    for node in nodes:
        visited[node.id] = False

    def dfs_helper(graph, node, cc):
        visited[node.id] = True
        cc.add_node(node)
        for node_out in graph.outgoing(node):
            if node_out.id in visited:
                if not visited[node_out.id]:
                    dfs_helper(graph, node_out, cc)
        for node_in in graph.incoming(node):
            if node_in.id in visited:
                if not visited[node_in.id]:
                    dfs_helper(graph, node_in, cc)

    for node in nodes:
        if not visited[node.id]:
            connected_component = ConnectedComponent()
            dfs_helper(graph, node, connected_component)
            connected_components.append(connected_component)
    return connected_components

def grow_connected_components(graph, connected_components):
    print("grow_non_stem_connected_components")
    for connected_component in connected_components:
        grow_connected_component(graph, connected_component)
    return connected_components

def grow_connected_component(graph, connected_component):
    visited = {}
    for node_id in graph.nodes:
        visited[node_id] = False

    new_nodes = list()
    def dfs_helper(graph, node):
        if node.is_stem() or node.is_concat(axis=1):
            new_nodes.append(node)
            return 
        visited[node.id] = True
        for node_in in graph.incoming(node):
            if not visited[node_in.id]:
                dfs_helper(graph, node_in)

    for node in connected_component.nodes.values():
        if not visited[node.id]:
            dfs_helper(graph, node)

    connected_component.add_nodes(new_nodes)

def merge_connected_components(connected_components):
    # Merge connected components if they share intersected nodes
    pool = set(connected_components)
    merged_ccs = []
    while pool:
        merged_ccs.append(pool.pop())
        while True:
            for cc in pool:
                if merged_ccs[-1].nodes.keys() & cc.nodes.keys():
                    merged_ccs[-1].merge(cc)
                    pool.remove(cc)
                    break
            else:
                break
    return merged_ccs

def get_remaining_nodes(connected_components, all_nodes):
    remaining_nodes = []
    included_nodes = []
    for cc in connected_components:
        included_nodes.extend([node.id for node in cc.nodes.values()])
    remaining_nodes = [all_nodes[node_id] for node_id in all_nodes if node_id not in included_nodes]
    return remaining_nodes
    
def group_individual_nodes(individual_nodes):
    print("group_individual_nodes")
    individual_connected_components = list()
    for node in individual_nodes:
        connected_component = ConnectedComponent()
        connected_component.add_node(node)
        individual_connected_components.append(connected_component)
    return individual_connected_components

def automated_partition_zigs(graph, opt=None):    
    # Step 1: Get non-stem nodes with shape dependent
    non_stem_nodes = get_non_stem_nodes(graph)    

    # Step 2: Find the connected components over non-stem nodes
    non_stem_connected_components = get_connected_components(graph, non_stem_nodes)
    
    # Step 3: Grow the connected components till all incoming nodes are stem nodes 
    # and all outgoing nodes has non-stem nodes but has stem outgoing nodes.
    # TODO: if no stem component anscenter is found, then the cc needs to be dropped off, because it is directly connected to input.
    grown_connected_components = grow_connected_components(graph, non_stem_connected_components)

    # Step 4: Merge connected components if any intersection
    merged_connected_components = merge_connected_components(grown_connected_components)

    # Step 5: Group the remaining parameters over individual nodes
    remaining_nodes = get_remaining_nodes(merged_connected_components, graph.nodes)
    individual_connected_components = group_individual_nodes(remaining_nodes)


    # Step 6: Setup connected components fo graph
    graph.set_connected_components(merged_connected_components + individual_connected_components)

    # Step 7: Set ZIGs type
    graph.set_zigs(opt)

    return graph



from .transforms import Rename, ConvBNFuse

# PyTorch Graph Transforms
FRAMEWORK_TRANSFORMS = [
    Rename(op=r"onnx::(.*)", to=r"\1"),
    Rename(op=r"gemm", to=r"linear"),
    Rename(op=r"batchnormalization", to="batchnorm"),
]

CONV_BN_FUSE = ConvBNFuse("conv > batchnorm", "convbn")



import imp
import re
import copy

from . import ge
import os
import sys
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

# from graph.node import Node

class Rename():
    def __init__(self, op=None, name=None, to=None):
        assert op or name, "Either op or name must be provided"
        assert not(op and name), "Either op or name should be provided, but not both"
        assert bool(to), "The to parameter is required" 
        self.to = to
        self.op = re.compile(op) if op else None
        self.name = re.compile(name) if name else None
    
    def apply(self, graph):
        for i, node in enumerate(graph.nodes.values()):
            if self.op:
                node.op.name = self.op.sub(self.to, node.op.name)
            if self.name is None:
                node.name = str(node.op.name)
            else:
                node.name = self.name.sub(self.to, node.name)


# class Fold():
#     def __init__(self, pattern, to, name=None):
#         # TODO: validate that op and name are valid
#         self.pattern = ge.GEParser(pattern).parse()
#         self.to = to
#         self.name = name

#     def apply(self, graph):     
#         while True:
#             matches, _ = graph.search(self.pattern)
#             if not matches:
#                 break

#             # Replace pattern with new node
#             if self.to == "__first__":
#                 combo = matches[0]
#             elif self.to == "__last__":
#                 combo = matches[-1]
#             else:
#                 # find the most bottom child
#                 outputs = set()
#                 match_ids = [node.id for node in matches]
#                 for match_node in matches:
#                     for outgoing_node in graph.outgoing(match_node):
#                         if outgoing_node.id not in match_ids:
#                             outputs.add(outgoing_node)
#                 # combine operators
#                 combo_op = matches[0].op
#                 for i in range(1, len(matches)):
#                     combo_op += matches[i].op
#                 combo_op.name = self.to or self.pattern
#                 combo = Node(id=graph.sequence_id(),
#                              op=combo_op,
#                              output_shape=matches[-1].output_shape,
#                              outputs = list(outputs)) # TODO, check bugs
#                 combo._caption = "/".join(filter(None, [l.caption for l in matches]))
#             graph.replace(matches, combo)


class ConvBNFuse():
    def __init__(self, pattern, to, name=None):
        self.pattern = ge.GEParser(pattern).parse()
        self.to = to
        self.name = name

    def apply(self, graph):     
        graph.fused_conv_bns = list()
        while True:
            matches, _ = graph.search(self.pattern)
            if not matches:
                break
            for match_node in matches:
                match_node._skip_pattern_search = True
            graph.fused_conv_bns.append(matches)


import torch
import numpy as np
import os
import sys
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)
from assets.group_types import GROUP_TYPE

class ConnectedComponent:
    def __init__(self):
        self.nodes = dict()
        self.params = set()
        self.type = 1
        self.epsilon = 0.0
        self.upper_group_sparsity = 0.0
        self.num_zero_groups = 0
        self.zero_groups_idxes = list()
        self.non_zero_group_idxes = list()
        self.onnx_params = set()
        self.num_groups = 0
        self.auxiliary_ccs = list()
        
    def __repr__(self):
        return f"Id: {self.id}; Group type: {self.type}; Node ids: {self.nodes.values()}; \
        Param: {self.params}; OnnxParam: {self.onnx_params}; Num groups: {self.num_groups}, Num zero groups: {self.num_zero_groups}"

    @property
    def id(self):
        return "_".join([node.id for node in self.nodes.values()])

    def add_node(self, node):
        self.nodes[node.id] = node
    
    def add_nodes(self, nodes):
        for node in nodes:
            if node.id not in self.nodes:
                self.nodes[node.id] = node
    
    def merge(self, other_cc):
        for node_id in other_cc.nodes:
            if node_id not in self.nodes:
                self.nodes[node_id] = other_cc.nodes[node_id]

    def set_properties(self, output_nodes, opt=None):
        self.type = GROUP_TYPE["default"]
        for node in self.nodes.values():
            if node.is_concat(axis=1):
                self.type = GROUP_TYPE["auxilary"]
            if node.op.name == "conv":
                self.type = GROUP_TYPE["conv"]
            elif node.op.name == "linear" or node.op.name == "gemm":
                self.type = GROUP_TYPE["linear"]
            elif node.op.name == "multi-head-linear":
                self.type = GROUP_TYPE["multi-head-linear"]
            self.params |= set(node.params)
        self.type = GROUP_TYPE["default"] if (self.contain_output_nodes(output_nodes) \
                    or self.contain_non_zero_invariant_nodes()) else self.type
        self.epsilon = opt["optimizer"]["epsilon"][self.type] if opt is not None else 0.0
        self.upper_group_sparsity = opt["optimizer"]["upper_group_sparsity"][self.type] if opt is not None else 1.0

    def get_zero_groups(self):
        xs = []
        for param in self.params:
            if len(param.data.shape) == 1:
                xs.append(param.data.unsqueeze(1))
            elif len(param.data.shape) == 4: # conv layer
                xs.append(param.data.view(param.data.shape[0], -1))
            else:
                xs.append(param.data)        
        flatten_x = torch.cat(xs, dim = 1)

        norm_groups = torch.norm(flatten_x, dim=1)
        zero_groups_idxes = norm_groups == 0
        nonzero_groups_idxes = norm_groups != 0
        self.num_zero_groups = int(torch.sum(zero_groups_idxes).item())
        self.zero_groups_idxes = np.arange(0, self.num_groups)[zero_groups_idxes]
        self.non_zero_group_idxes = np.arange(0, self.num_groups)[nonzero_groups_idxes]

    def contain_concat_node(self, axis=1):  
        for node in self.nodes.values():
            if node.is_concat(axis=axis):
                return True
        return False
    
    def is_auxilary(self):
        if self.contain_concat_node(axis=1):
            self.type = GROUP_TYPE["auxilary"]
            return True
        else:
            return False
    
    def get_concat_node(self):
        # One auxiliary connected component has at most one concat joint node 
        # based on the algorithm
        if not self.is_auxilary():
            return None
        else:
            for node in self.nodes.values():
                if node.is_concat(axis=1):
                    return node
            return None
    
    def contain_output_nodes(self, output_nodes):
        for node in self.nodes.values():
            if node.id in output_nodes:
                return True
        return False

    def contain_non_zero_invariant_nodes(self):
        for node in self.nodes.values():
            if not node.is_zero_invariant():
                return True
        return False


import imp
import torch
import numpy as np
from collections import defaultdict

from .connected_component import GROUP_TYPE
from .node import Node
from assets.theme import THEMES
from operation.operator import OP_DICT, Operator
from transform.onnx_graph_transform import FRAMEWORK_TRANSFORMS, CONV_BN_FUSE
from distutils.version import LooseVersion

class Graph():
    """Tracks nodes and edges of a directed graph and supports basic operations on them."""
    def __init__(self, model=None, dummy_input=None):
        print("graph constructor")
        self.nodes = dict()
        self.edges = list()
        self.connected_components = dict()
        self.output_nodes = list()
        self.paths = list()
        self.theme = THEMES["basic"]
        # parameters of neural network model with names 
        self.params_grad = dict()
        self.params_no_grad = dict()
        self.params_names = list()
        self.params_onnx = dict()
        self.params_to_nodes = dict()
        self.inputs = dict()

        for name, param in model.named_parameters():
            if param.requires_grad:
                self.params_grad[name] = param
        
        for name in model.state_dict():
            self.params_names.append(name)
            if name not in self.params_grad:
                self.params_no_grad[name] = model.state_dict()[name]

        if model:
            assert dummy_input is not None, "Dummy_input args must be provided for Pytorch models."
            model = model.eval()
            self.build(model, dummy_input)

            # Apply Transforms
            for t in FRAMEWORK_TRANSFORMS:
                t.apply(self)
            CONV_BN_FUSE.apply(self)

    def id(self, node):
        """Returns a unique node identifier. If the node has an id
        attribute (preferred), it's used. Otherwise, the hash() is returned."""
        return node.id if hasattr(node, "id") else hash(node)

    def add_node(self, node):
        id = self.id(node)
        if len(self.nodes) == 0:
            self.root_id = id
        self.nodes[id] = node

    def root(self):
        try:
            return self.nodes[self.root_id]
        except:
            return None

    def add_edge(self, node1, node2, label=None):
        # If the edge is already present, don't add it again.
        # TODO: If an edge exists with a different label, still don't add it again.
        edge = (self.id(node1), self.id(node2), label)
        if edge not in self.edges:
            self.edges.append(edge)

    def add_edge_by_id(self, vid1, vid2, label=None):
        self.edges.append((vid1, vid2, label))

    def outgoing(self, node):
        """Returns nodes connecting out of the given node (or list of nodes)."""
        nodes = node if isinstance(node, list) else [node]
        node_ids = [self.id(n) for n in nodes]
        # Find edges outgoing from this group but not incoming to it
        outgoing = [self[e[1]] for e in self.edges
                    if e[0] in node_ids and e[1] not in node_ids]
        return outgoing

    def incoming(self, node):
        """Returns nodes connecting to the given node (or list of nodes)."""
        nodes = node if isinstance(node, list) else [node]
        node_ids = [self.id(n) for n in nodes]
        # Find edges incoming to this group but not outgoing from it
        incoming = [self[e[0]] for e in self.edges
                    if e[1] in node_ids and e[0] not in node_ids]
        return incoming

    def siblings(self, node):
        """Returns all nodes that share the same parent (incoming node) with
        the given node, including the node itself.
        """
        incoming = self.incoming(node)
        # TODO: Not handling the case of multiple incoming nodes yet
        if len(incoming) == 1:
            incoming = incoming[0]
            siblings = self.outgoing(incoming)
            return siblings
        else:
            return [node]

    def __getitem__(self, key):
        if isinstance(key, list):
            return [self.nodes.get(k) for k in key]
        else:
            return self.nodes.get(key)

    def remove(self, nodes):
        """Remove a node and its edges."""
        nodes = nodes if isinstance(nodes, list) else [nodes]
        for node in nodes:
            k = self.id(node)
            self.edges = list(filter(lambda e: e[0] != k and e[1] != k, self.edges))
            del self.nodes[k]
    
    def set_connected_components(self, connected_components):
        for cc in connected_components:
            self.connected_components[cc.id] = cc
            for node in cc.nodes.values():
                node.connected_components[cc.id] = cc

    def print(self):
        for node_id in self.nodes:
            print(node_id, self.nodes[node_id])

        for edge in self.edges:
            print(edge)

        for tensor_key in self.tensors:
            print(tensor_key, self.tensors[tensor_key])

    def _parse_tensors_info(self, state_dict, torch_graph_str):
        """Use hack to parse tensor info, should be better option for doing it"""
        prefix_str = "graph"
        assert torch_graph_str.startswith(prefix_str), "Invalid graph str to be parsed"

        tensors_str = get_str_inside_parenthesis(torch_graph_str, prefix_str = prefix_str)
        tensors_str_list = [s.strip() for s in tensors_str.split('%')][1:]

        self.params_id_to_name = dict()

        num_inputs = len(tensors_str_list) - len(state_dict)
        cur_input = 0
        cur_param = 0
        for i, tensor_str in enumerate(tensors_str_list):
            tensor_str_split = [s.strip() for s in tensor_str.split(":")]
            tensor_id = tensor_str_split[0]
            tensor_info = [s.strip() for s in tensor_str_split[1].split(",")]
            required_grad = True if [s for s in tensor_info if "requires_grad" in s][0].split('=')[1] == '1' else False
            tensor_type = "input" if i < num_inputs else "params"
            if tensor_type == "input":
                self.inputs[cur_input] = (tensor_id, tensor_str_split)
                cur_input += 1
            elif tensor_type == "params":
                self.params_id_to_name[int(tensor_id)] = self.params_names[cur_param]
                cur_param += 1

    def get_output_shape(self, torch_node_str):
        str_to_processed = torch_node_str.split(':')[1].strip()
        output_str = get_str_inside_parenthesis(str_to_processed, prefix_str = "Float")
        if output_str is None:
            return None
        output_str_splits = output_str.split(',')
        output_shapes = []
        for item in output_str_splits:
            item = item.strip()
            if item.isnumeric():
                output_shapes.append(int(item))
            else:
                break
        return output_shapes

    def build(self, model, dummy_input=None):
        # Run the Pytorch graph to get a trace and generate a graph from it
        trace_graph, _ = torch.jit._get_trace_graph(model, dummy_input)
        torch_graph = None
        if LooseVersion(torch.__version__) >= LooseVersion('1.9.0') and \
           LooseVersion(torch.__version__) <= LooseVersion('1.11.10'):
            torch_graph = torch.onnx._optimize_trace(trace_graph, torch.onnx.OperatorExportTypes.ONNX)
        elif LooseVersion(torch.__version__) >= LooseVersion('1.13.0'):
            torch_graph = torch.onnx._optimize_graph(trace_graph, torch.onnx.OperatorExportTypes.ONNX)
        else:
            raise "Torch {} is not supported because of some bug in _optimize_trace.".format(torch.__version__)

        # TODO: Should be better way to get the information of tensors from torch_graph
        self._parse_tensors_info(model.state_dict(), str(torch_graph))

        # Loop through nodes from torch_graph to build graph for OTO
        torch_nodes_by_inputs = defaultdict(set)
        torch_nodes_by_outputs = defaultdict(set)
        for torch_node in torch_graph.nodes():
            # Get Operation
            op_name = torch_node.kind().split("::")[-1].lower()
            # Operation Parameters
            op_params = {k: getattr(torch_node, torch_node.kindOf(k))(k) for k in torch_node.attributeNames()}
            op = None
            if op_name in OP_DICT:
                op= OP_DICT[op_name]
            else:
                print('Unknown op:', op_name)
                op = Operator(name=op_name, params=op_params)
                OP_DICT[op_name] = op

            # Inputs/outputs
            inputs = [i.unique() for i in torch_node.inputs()]
            outputs = [o.unique() for o in torch_node.outputs()]

            params = [self.params_id_to_name[i] for i in inputs if i in self.params_id_to_name]
            # Only consider non-params inputs
            inputs = [i for i in inputs if i not in self.params_id_to_name]
 
            output_shape = self.get_output_shape(str(torch_node))

            # Add nodes
            node = Node(id=pytorch_id(torch_node), op=op, op_params=op_params, \
                        params=params, inputs=inputs, outputs=outputs, output_shape=output_shape)

            self.add_node(node)

            # Add edges
            for output in outputs:
                torch_nodes_by_outputs[output].add(torch_node)
                for target_torch_node in torch_nodes_by_inputs[output]:
                    self.add_edge_by_id(pytorch_id(torch_node), pytorch_id(target_torch_node))
            for input in inputs:
                torch_nodes_by_inputs[input].add(torch_node)
                for target_torch_node in torch_nodes_by_outputs[input]:
                    self.add_edge_by_id(pytorch_id(target_torch_node), pytorch_id(torch_node))

        for node in self.nodes.values():
            if len(self.outgoing(node)) == 0:
                self.output_nodes.append(node.id)

        for i, node in enumerate(self.nodes.values()):
            if len(node.inputs) == 0:
                continue
            nodes_in = self.incoming(node)
            if len(nodes_in) == 0:
                for in_id in node.inputs:
                    in_id = in_id[4:] #skip out- prefix
                    input_tensor_info = self.inputs[int(in_id)]
                    tensor_shape_str = get_str_inside_parenthesis(input_tensor_info[1][1], prefix_str='Float')
                    input_shape = []
                    if tensor_shape_str:
                        tensor_shape_str_splits = tensor_shape_str.split(',')
                        for item in tensor_shape_str_splits:
                            item = item.strip()
                            if item.isnumeric():
                                input_shape.append(int(item))
                            else:
                                break
                    node.input_shape.append(input_shape)
            else:
                for node_in in nodes_in:
                    node.input_shape.append(node_in.output_shape)

    def compress(self):
        import onnx
        from onnx import numpy_helper
        # Each CC prunes row-wisely based on non-zero groups.
        for cc in self.connected_components.values():
            for node in cc.nodes.values():
                node.pruned_shape = [-1] * 2
                if len(cc.non_zero_group_idxes) > 0:
                    node.pruned_shape[0] = len(cc.non_zero_group_idxes)
                    if hasattr(cc, 'raw_num_groups'):
                        if cc.raw_num_groups != cc.num_groups:
                            node.pruned_shape[0] = len(cc.raw_non_zero_group_idxes)
            if cc.type <= GROUP_TYPE["auxilary"]:
                continue
            for name in cc.onnx_params:
                numpy_param = onnx.numpy_helper.to_array(self.params_onnx[name])
                if cc.raw_num_groups == cc.num_groups:
                    pruned_onnx_param = numpy_param[cc.non_zero_group_idxes, ...]
                else:
                    pruned_onnx_param = numpy_param[cc.raw_non_zero_group_idxes, ...]

                onnx_param = onnx.helper.make_tensor(
                        name=name,
                        data_type=onnx.TensorProto.FLOAT,
                        dims=pruned_onnx_param.shape,
                        vals=pruned_onnx_param.flatten().tolist())
                self.params_onnx[name].CopyFrom(onnx_param)

        def dfs_helper(graph, node, cc_id, visited, incoming_cc_ids):
            if node.cc_id != cc_id:
                incoming_cc_ids.add(node.cc_id)
                return 
            visited[node.id] = True
            for node_in in graph.incoming(node):
                if not visited[node_in.id]:                    
                    dfs_helper(graph, node_in, cc_id, visited, incoming_cc_ids)
        
        # Each node prunes redundancy depending on incoming or dependent CC.
        for cc in self.connected_components.values():
            if cc.is_auxilary():
                # print(cc)
                for name in cc.onnx_params:
                    # print(name)
                    node = self.params_to_nodes[name][0]
                    numpy_param = onnx.numpy_helper.to_array(self.params_onnx[name])
                    pruned_onnx_param = numpy_param[cc.non_zero_group_idxes, ...]
                    onnx_param = onnx.helper.make_tensor(
                            name=name,
                            data_type=onnx.TensorProto.FLOAT,
                            dims=pruned_onnx_param.shape,
                            vals=pruned_onnx_param.flatten().tolist())
                    self.params_onnx[name].CopyFrom(onnx_param)                     

            for name in cc.onnx_params:
                # TODO: Weight sharing, i.e., one param has multiple nodes
                node = self.params_to_nodes[name][0]
                numpy_param = onnx.numpy_helper.to_array(self.params_onnx[name])
                visited = {}
                for node_id in self.nodes:
                    visited[node_id] = False
                incoming_cc_ids = set()
                dfs_helper(self, node, node.cc_id, visited, incoming_cc_ids)
                # If no-incoming CCs, then the node is directly connected regardless stem node to the input
                # there is no need to adjust parameters
                if len(incoming_cc_ids) == 0:
                    continue
                else:
                    incoming_cc = self.connected_components[next(iter(incoming_cc_ids))]
                    if incoming_cc.type <= GROUP_TYPE['default']:
                        continue
                    if len(numpy_param.shape) >= 2 and (node.is_conv() or node.is_linear()):
                        node.pruned_shape[1] = len(incoming_cc.non_zero_group_idxes)
                        pruned_onnx_param = numpy_param[:, incoming_cc.non_zero_group_idxes, ...]
                        onnx_param = onnx.helper.make_tensor(
                            name=name,
                            data_type=onnx.TensorProto.FLOAT,
                            dims=pruned_onnx_param.shape,
                            vals=pruned_onnx_param.flatten().tolist())
                        self.params_onnx[name].CopyFrom(onnx_param)

    def search(self, pattern):
        """Searches the graph for a sub-graph that matches the given pattern
        and returns the first match it finds.
        """
        for node in self.nodes.values():
            match, following = pattern.match(self, node)
            if match:
                return match, following
        return [], None

    def sequence_id(self):
        from random import getrandbits
        return getrandbits(64)

    def replace(self, nodes, node):
        """Replace nodes with node. Edges incoming to nodes[0] are connected to
        the new node, and nodes outgoing from nodes[-1] become outgoing from
        the new node."""
        nodes = nodes if isinstance(nodes, list) else [nodes]
        # Is the new node part of the replace nodes (i.e. want to collapse
        # a group of nodes into one of them)?
        collapse = self.id(node) in self.nodes
        # Add new node and edges
        if not collapse:
            self.add_node(node)
        for in_node in self.incoming(nodes):
            # TODO: check specifically for output_shape is not generic. Consider refactoring.
            self.add_edge(in_node, node, in_node.output_shape if hasattr(in_node, "output_shape") else None)
        for out_node in self.outgoing(nodes):
            self.add_edge(node, out_node, node.output_shape if hasattr(node, "output_shape") else None)
        # Remove the old nodes
        for n in nodes:
            if collapse and n == node:
                continue
            self.remove(n)
    
    def set_zigs(self, opt=None):
        self.auxilary_ccs = dict()
        # First pass to set up stem ccs
        for cc_id in self.connected_components:
            cc = self.connected_components[cc_id]
            cc.set_properties(self.output_nodes, opt)
            if not cc.is_auxilary():
                # If cc is stem but no params_grad, then all params are not required grad
                shapes_grad_params = [self.params_grad[name].shape[0] for name in cc.params if name in self.params_grad]
                if len(shapes_grad_params) > 0:
                    cc.num_groups = max(shapes_grad_params)
                else:
                    shapes_nograd_params = [self.params_no_grad[name].shape[0] for name in cc.params if name in self.params_no_grad]
                    cc.num_groups = max(shapes_nograd_params) if len(shapes_nograd_params) > 0 else 0
                    cc.type = GROUP_TYPE['no-update']
            else:
                self.auxilary_ccs[cc_id] = cc
                cc.dependent_stem_ccs = list()

        # Second pass to tackle auxiliary connected components
        visited = dict()
        for cc_id in self.auxilary_ccs:
            visited[cc_id] = False
        
        def dfs_helper(graph, cc, dependent_stem_ccs):
            if not cc.is_auxilary():
                dependent_stem_ccs.append(cc.id)
                return 
            elif visited[cc.id]:
                dependent_stem_ccs.extend(cc.dependent_stem_ccs)
                return 

            # if cc is auxiliary, get its concat node if any
            concat_node = cc.get_concat_node()
            if concat_node is None:
                return 

            for node_in_id in concat_node.inputs:
                node_in = graph.nodes[node_in_id]
                cc_in = graph.connected_components[node_in.cc_id]
                dfs_helper(self, cc_in, dependent_stem_ccs)

        for auxilary_cc in self.auxilary_ccs.values():
            if visited[auxilary_cc.id] is True:
                continue
            dfs_helper(self, auxilary_cc, auxilary_cc.dependent_stem_ccs)
            visited[auxilary_cc.id] = True
    
        # Third pass for tackling connection between stem cc and auxiliary ccs
        for cc in self.connected_components.values():
            if cc.is_auxilary():
                if len(cc.params) == 0:
                    continue
                offset = 0
                for depend_cc_id in cc.dependent_stem_ccs:
                    depend_cc = self.connected_components[depend_cc_id]
                    depend_cc.auxiliary_ccs.append((cc.id, offset))
                    offset += depend_cc.num_groups

    def random_set_zero_groups(self, group_sparsity=0.5):
        for cc in self.connected_components.values():
            if cc.is_auxilary():
                continue
            # if cc.type is not stem, or default with no group sparsity penalize, skip
            if cc.type <= GROUP_TYPE["auxilary"]:
                continue
            num_zero_groups = np.random.randint(1, cc.num_groups - 1)
            # num_zero_groups = max(max(1, int((group_sparsity + np.random.random()) * cc.num_groups - 1)), cc.num_groups - 1)
            # num_zero_groups = int(group_sparsity * cc.num_groups - 1)
            zero_group_idxes = np.random.choice(list(range(0, cc.num_groups - 1)), num_zero_groups, replace=False)
            cc_params_objs = [self.params_grad[name] if name in self.params_grad else self.params_no_grad[name] for name in cc.params]
            for p in cc_params_objs:
                p.data[zero_group_idxes, ...] = 0.0
            
            for (auxiliary_cc_id, offset) in cc.auxiliary_ccs:
                auxiliary_cc = self.connected_components[auxiliary_cc_id]
                auxiliary_cc_params_objs = [self.params_grad[name] if name in self.params_grad else self.params_no_grad[name] for name in auxiliary_cc.params]
                for p in auxiliary_cc_params_objs:
                    p.data[zero_group_idxes + offset, ...] = 0.0

    def set_zero_groups(self):
        for cc in self.connected_components.values():
            if cc.is_auxilary():
                continue
            cc_params_objs = [self.params_grad[name] if name in self.params_grad else self.params_no_grad[name] for name in cc.params]
            xs = []
            cc.raw_num_groups = cc.num_groups
            for param in cc_params_objs:
                cc.raw_num_groups = param.shape[0]     
                xs.append(param.data.view(cc.num_groups, -1))

            for (auxiliary_cc_id, offset) in cc.auxiliary_ccs:
                auxiliary_cc = self.connected_components[auxiliary_cc_id]
                auxiliary_cc_params_objs = [(name, self.params_grad[name]) if name in self.params_grad else (name, self.params_no_grad[name]) for name in auxiliary_cc.params]
                for name, param in auxiliary_cc_params_objs:
                    if len(param.data.shape) == 1:
                        xs.append(param.data[offset:offset + cc.num_groups,...].unsqueeze(1))
                    else:
                        xs.append(param.data[offset:offset + cc.num_groups,...].view(cc.num_groups, -1))
            if len(xs) == 0:
                continue            
            flatten_x = torch.cat(xs, dim = 1)
            norm_groups = torch.norm(flatten_x, dim=1)
            zero_groups_idxes = norm_groups == 0
            nonzero_groups_idxes = norm_groups != 0
            cc.num_zero_groups = int(torch.sum(zero_groups_idxes).item())
            cc.zero_groups_idxes = np.arange(0, cc.num_groups)[zero_groups_idxes.cpu()]
            cc.non_zero_group_idxes = np.arange(0, cc.num_groups)[nonzero_groups_idxes.cpu()]

            if cc.raw_num_groups != cc.num_groups:
                repeat_time = cc.raw_num_groups // cc.num_groups
                zero_groups_idxes = zero_groups_idxes.repeat_interleave(repeat_time)
                nonzero_groups_idxes = nonzero_groups_idxes.repeat_interleave(repeat_time)
            cc.raw_zero_groups_idxes = np.arange(0, cc.raw_num_groups)[zero_groups_idxes.cpu()]
            cc.raw_non_zero_group_idxes = np.arange(0, cc.raw_num_groups)[nonzero_groups_idxes.cpu()]

        for cc in self.connected_components.values():
            if cc.is_auxilary():
                non_zero_group_idxes = []
                offset = 0
                for dependent_stem_cc_id in cc.dependent_stem_ccs:
                    dependent_stem_cc = self.connected_components[dependent_stem_cc_id]
                    non_zero_group_idxes.append(dependent_stem_cc.non_zero_group_idxes + offset)
                    offset += dependent_stem_cc.num_groups
                cc.non_zero_group_idxes = np.concatenate(non_zero_group_idxes)

    def build_dot(self, verbose=True, vertical=False):
        """
        Generate a GraphViz Dot graph.
        If verbose, then draw more detailed info as well as groups.
        Returns a GraphViz Digraph object.
        """
        from graphviz import Digraph
        import random

        dot = Digraph()
        dot.attr("graph", 
                bgcolor=self.theme["background_color"],
                color=self.theme["outline_color"],
                fontsize=self.theme["font_size"],
                fontcolor=self.theme["font_color"],
                fontname=self.theme["font_name"],
                margin=self.theme["margin"],
                rankdir="TB" if vertical else "LR",
                pad=self.theme["padding"])

        dot.attr("edge", style="solid", 
                color=self.theme["outline_color"],
                fontsize=self.theme["font_size"],
                fontcolor=self.theme["font_color"],
                fontname=self.theme["font_name"])

        # Build GraphViz Digraph
        if not verbose or len(self.connected_components) == 0:
            dot.attr("node", shape="box", 
                    style="filled", margin="0,0",
                    fillcolor=self.theme["fill_color"],
                    color=self.theme["outline_color"],
                    fontsize=self.theme["font_size"],
                    fontcolor=self.theme["font_color"],
                    fontname=self.theme["font_name"])

            for node in self.nodes.values():
                label = "<tr><td cellpadding='6'>{}</td></tr>".format(node.title)
                label = "<<table border='0' cellborder='0' cellpadding='0'>" + label + "</table>>"
                dot.node(str(node.id), label)
        else:
            for cc in self.connected_components.values():
                random_number = random.randint(0,16777215)
                hex_number = str(hex(random_number))
                color ='#'+ hex_number[2:]
                dot.attr("node", shape="box", 
                        style="filled", margin="0,0",
                        fillcolor=color,
                        color=color,
                        fontsize=self.theme["font_size"],
                        fontcolor="#FFFFFF",
                        fontname=self.theme["font_name"])
                for node in cc.nodes.values():
                    label = "<tr><td cellpadding='6'>{}</td></tr>".format(node.title)
                    if node.id:
                        label += "<tr><td>{}</td></tr>".format(node.id)
                    label = "<<table border='0' cellborder='0' cellpadding='0'>" + label + "</table>>"
                    dot.node(str(node.id), label)                        
        for a, b, label in self.edges:
            if isinstance(label, (list, tuple)):
                label = "x".join([str(l or "?") for l in label])
            dot.edge(str(a), str(b), label)
        return dot    
        
    def params_groups(self, epsilon=[]):            
        param_groups = dict()
        
        for cc in self.connected_components.values():
            cc_param_groups = dict()
            cc.params_grad = cc.params & self.params_grad.keys()
            cc_param_groups['cc_id'] = cc.id
            cc_param_groups['params'] = [self.params_grad[name] for name in cc.params_grad] 
            cc_param_groups['group_type'] = cc.type
            if isinstance(epsilon, list):
                cc_param_groups['epsilon'] = cc.epsilon if len(epsilon) == 0 else epsilon[cc.type]
            else:
                cc_param_groups['epsilon'] = epsilon
            cc_param_groups['upper_group_sparsity'] = cc.upper_group_sparsity
            cc_param_groups['names'] = cc.params_grad
            cc_param_groups['shapes'] = [self.params_grad[name].shape for name in cc.params_grad] 
            cc_param_groups['auxiliary_ccs'] = list()
            if len(cc_param_groups['params']) > 0:
                cc_param_groups['num_groups'] = cc.num_groups
                param_groups[cc.id] = cc_param_groups

        # Second pass for tackling auxliary cc
        for cc in self.connected_components.values():
            if cc.is_auxilary():
                if len(cc.params) == 0:
                    continue
                offset = 0
                for depend_cc_id in cc.dependent_stem_ccs:
                    depend_cc = self.connected_components[depend_cc_id]
                    depend_cc_param_groups = param_groups[depend_cc.id]
                    depend_cc_param_groups['auxiliary_ccs'].append((cc.id, offset))
                    offset += depend_cc.num_groups
        param_groups = dict(sorted(param_groups.items(), key=lambda kv:(kv[0], kv[1])))
        return param_groups.values()
    
def pytorch_id(node):
    """Returns a unique ID for a node."""
    return "out-" + "-".join(["{}".format(o.unique()) for o in node.outputs()])

def get_str_inside_parenthesis(str_to_processed, prefix_str=None):
    if not str_to_processed.startswith(prefix_str):
        return None
    stack = []
    start_idx = len(prefix_str) + 1
    end_idx = -1 
    for c in str_to_processed:
        if c == '(':
            stack.append(c)
        elif c == ')':
            stack.pop()
        end_idx += 1
        if len(stack) == 0 and end_idx > len(prefix_str):
            break
    return str_to_processed[start_idx : end_idx] 



import numpy as np
import os
import sys
currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)

from operation.operator import Operator

class Node():
    """Represents a framework-agnostic neural network layer in a directed graph."""

    def __init__(self, id, op, op_params=None, params=[], inputs=[], outputs=[], output_shape=[]):
        """
        id: unique ID for the layer that doesn't repeat in the computation graph.
        name: Name to display
        op: Framework-agnostic operation .
        """
        self.id = id
        self.connected_components = dict() # the connected component id that it belongs to.
        if isinstance(op, Operator):
            self.op = op
        else:
            self.op = Operator(name=op, params=op_params)
        
        self.inputs = ['out-' + str(i) for i in inputs]
        self.outputs = ['out-' + str(o) for o in outputs]
        self.params = params    
        self.output_shape = output_shape
        self.input_shape = []
        self.op_params = op_params if op_params else {}
        self._skip_pattern_search = False
        
    @property
    def cc_id(self):
        return list(self.connected_components.keys())[0]

    @property
    def title(self):
        # Default
        title = self.op.name
        if "kernel_shape" in self.op_params:
            # Kernel
            kernel = self.op_params["kernel_shape"]
            title += "x".join(map(str, kernel))
        if "stride" in self.op_params:
            stride = self.op_params["stride"]
            if np.unique(stride).size == 1:
                stride = stride[0]
            if stride != 1:
                title += "/s{}".format(str(stride))
        return title

    def __repr__(self):
        args = (self.op, self.id, self.title, self.inputs, self.outputs)
        f = "<Node: op: {}, id: {}, title: {}, inputs: {}, outputs: {}"
        if self.params:
            args += (str(self.params),)
            f += ", params: {:}"
        if self.op_params:
            args += (str(self.op_params),)
            f += ", op_params: {:}"
        if self.output_shape:
            args += (str(self.output_shape),)
            f += ", output_shape: {:}"
        if self.input_shape:
            args += (str(self.input_shape),)
            f += ", input_shape: {:}"
        f += ">"
        return f.format(*args)

    def is_concat(self, axis=None):
        # Check if concat at first
        _is_concat = self.op.name == "Concat" or self.op.name == 'concat'
        if axis == None:
            return _is_concat
        # Check if axis match
        if _is_concat:
            if 'axis' in self.op_params:
                if self.op_params['axis'] == axis:
                    return True
                else:
                    return False
            else:
                return False
        return _is_concat

    def is_conv(self):
        return self.op.name == "Conv" or self.op.name == 'conv'

    def is_linear(self):
        return self.op.name == "Linear" or self.op.name == 'linear' \
            or self.op.name == "Gemm" or self.op.name == "gemm"

    def is_stem(self):
        return self.op.type == "Stem" or self.op.type == "stem"

    def is_zero_invariant(self):
        return self.op.zero_invariant
    



THEMES = {
    "basic": {
        "background_color": "#FFFFFF",
        "fill_color": "#E8E8E8",
        "outline_color": "#000000",
        "font_color": "#000000",
        "font_name": "Times",
        "font_size": "10",
        "margin": "0,0",
        "padding":  "1.0,0.5",
    },
}


import numpy as np
import torch
import torch.nn as nn
import copy

MILLION = 1e6

def compute_linear_flops(node, compressed=False):
    if not node.input_shape or not node.output_shape:
        return 0
    output_shape = node.output_shape
    input_shape = node.input_shape[0]
    if compressed:
        pruned_shape = node.pruned_shape
        input_shape[1] = pruned_shape[1] if pruned_shape[1] >= 0 else input_shape[1]
        output_shape[1] = pruned_shape[0] if pruned_shape[0] >= 0 else output_shape[1]
    input_features = input_shape[1]
    output_features = output_shape[1]
    num_ops = input_features * output_features / float(MILLION)
    return num_ops

def compute_conv_flops(node, compressed=False):
    if not node.input_shape or not node.output_shape:
        return 0
    output_shape = copy.deepcopy(node.output_shape) # B * C * H * W
    input_shape = copy.deepcopy(node.input_shape[0]) # B * C * H * W
    if compressed:
        pruned_shape = node.pruned_shape
        input_shape[1] = pruned_shape[1] if pruned_shape[1] >= 0 else input_shape[1]
        output_shape[1] = pruned_shape[0] if pruned_shape[0] >= 0 else output_shape[1]
    output_area = output_shape[-1] * output_shape[-2]
    filter_area = np.prod(node.op_params["kernel_shape"])
    output_channel = output_shape[1]
    input_channel = input_shape[1]
    group = node.op_params["group"]
    num_ops = float(input_channel*output_channel*filter_area*output_area  // group) / float(MILLION)
    return num_ops

def compute_flops(graph, compressed=False):
    flops_break_down = dict()
    flops_break_down['total'] = 0
    for cc in graph.connected_components.values():
        flops_break_down[cc.id] = dict()
        flops_break_down[cc.id]['flops'] = 0
        flops_break_down[cc.id]['num_groups'] = cc.num_groups
        for node in cc.nodes.values():
            if node.is_conv():
                cur_flops = compute_conv_flops(node, compressed)
                flops_break_down[cc.id]['flops'] += cur_flops
                flops_break_down['total'] += cur_flops
            if node.is_linear():
                cur_flops = compute_linear_flops(node, compressed)
                flops_break_down[cc.id]['flops'] += cur_flops
                flops_break_down['total'] += cur_flops
                
    for cc in graph.connected_components.values():
        flops_break_down[cc.id]['flops'] /= flops_break_down['total']
    return flops_break_down


from time import sleep
import torch
import os

def automated_compression(oto_graph, model, dummy_input, compressed_model_path, dynamic_axes=[False, dict()]):
    compressed_model_path = './' if compressed_model_path is None else compressed_model_path
    full_group_sparse_model_path = os.path.join(compressed_model_path, (model.name if hasattr(model, 'name') else type(model).__name__) + "_full_group_sparse.onnx" )
    compressed_model_path = os.path.join(compressed_model_path, (model.name if hasattr(model, 'name') else type(model).__name__) + "_compressed.onnx" )
    device = next(model.parameters()).device

    if dynamic_axes[0]:
        torch.onnx.export(
            model, 
            dummy_input.to(device), 
            full_group_sparse_model_path, 
            input_names=['input'],
            output_names=['output'],
            dynamic_axes=dynamic_axes[1]
        )
    else:
        torch.onnx.export(
            model, 
            dummy_input.to(device), 
            full_group_sparse_model_path
        )  

    import onnx
    max_try_count = 10
    try_count = 0
    onnx_model = None
    while True:
        sleep(1)
        if os.path.isfile(full_group_sparse_model_path):
            onnx_model = onnx.load(full_group_sparse_model_path)
            break
        try_count += 1
        if try_count >= max_try_count:
            break
    
    onnx.checker.check_model(onnx_model)
    onnx_graph = onnx_model.graph
    
    oto_graph.set_zero_groups()
    assign_onnx_tensors_on_oto(oto_graph, onnx_graph)

    for param in oto_graph.params_onnx:
        if param not in oto_graph.params_to_nodes:
            oto_graph.params_to_nodes[param] = list()
        for node in oto_graph.nodes.values():
            if param in node.params:
                oto_graph.params_to_nodes[param].append(node)
    
    oto_graph.compress()

    onnx.checker.check_model(onnx_model)
    onnx.save(onnx_model, compressed_model_path)

    return onnx_model, compressed_model_path, full_group_sparse_model_path


def assign_onnx_tensors_on_oto(oto_graph, onnx_graph):
    # print("assign_onnx_tensors_on_oto")
    remaining_tensors = list()
    for i, tensor in enumerate(onnx_graph.initializer):
        # print(f"Tensor Name: {tensor.name}, Data Type: {tensor.data_type}, Shape: {tensor.dims}")
        matched = False
        for cc in oto_graph.connected_components.values():
            if tensor.name in cc.params:
                cc.onnx_params.add(tensor.name)
                oto_graph.params_onnx[tensor.name] = tensor  
                matched = True
        if not matched:
            remaining_tensors.append(tensor)

    # Tackle the Conv-BN fusion
    conv_param_idx = -1
    for tensor in remaining_tensors:
        if "Conv" not in tensor.name and "conv" not in tensor.name:
            continue
        # print(f"Tensor Name: {tensor.name}, Data Type: {tensor.data_type}, Shape: {tensor.dims}")
        if len(tensor.dims) == 4:
            conv_param_idx += 1
        conv_bn_fuse = oto_graph.fused_conv_bns[conv_param_idx]
        cc_id = conv_bn_fuse[0].cc_id
        cc = oto_graph.connected_components[cc_id]
        cc.onnx_params.add(tensor.name)
        oto_graph.params_onnx[tensor.name] = tensor
        
        for node in conv_bn_fuse:
            if node.is_conv():
                node.params.append(tensor.name)



from .operators_dict import OPERATOR_DICT

class Operator:
    def __init__(self, name, params=None, zero_invariant=False, type="", shape_dependent=False, op_dict=None):
        self.name = name.lower()
        self.zero_invariant = zero_invariant
        self.type = type
        self.shape_dependent = shape_dependent
        self.params = params if params else {}
        if op_dict is not None:
            self.zero_invariant = True if op_dict['Zero-Invariant'] == 'TRUE' else False
            self.type = op_dict['Type']
            self.shape_dependent = True if op_dict['Shape-Dependent'] == 'TRUE' else False

    def __repr__(self):
        return self.full_info()

    def full_info(self):
        return "Name: " + self.name + ", zero_invariant: " + str(self.zero_invariant) + ", type: " + self.type + \
              ", shape_dependent: " + str(self.shape_dependent)

    def __eq__(self, op_name):
        return self.name == op_name

    def __add__(self, *other_ops):
        new_name = self.name
        new_zero_invariant = self.zero_invariant
        new_type = self.type
        new_shape_dependent = self.shape_dependent
        for op in other_ops:
            new_name += '_'  + op.name
            new_zero_invariant &= op.zero_invariant
            new_type += '_' + op.type
            new_shape_dependent &= op.shape_dependent
        return Operator(name=new_name, zero_invariant=new_zero_invariant, type=new_type, shape_dependent=new_shape_dependent)

OP_DICT = {}
for op_name in OPERATOR_DICT:
    op_dict = OPERATOR_DICT[op_name]
    op_name = op_name.lower()
    op = Operator(name=op_name, op_dict=op_dict)
    OP_DICT[op_name] = op



"""
This file is borrowed from hiddenlayer
Licensed under the MIT License
"""

import re

class GEParser():
    def __init__(self, text):
        self.index = 0
        self.text = text

    def parse(self):
        return self.serial() or self.parallel() or self.expression()

    def parallel(self):
        index = self.index
        expressions = []
        while len(expressions) == 0 or self.token("|"):
            e = self.expression()
            if not e:
                break
            expressions.append(e)
        if len(expressions) >= 2:
            return ParallelPattern(expressions)
        # No match. Reset index
        self.index = index
    
    def serial(self):
        index = self.index
        expressions = []
        while len(expressions) == 0 or self.token(">"):
            e = self.expression()
            if not e:
                break
            expressions.append(e)

        if len(expressions) >= 2:
            return SerialPattern(expressions)
        self.index = index

    def expression(self):
        index = self.index
        
        if self.token("("):
            e = self.serial() or self.parallel() or self.op()
            if e and self.token(")"):
                return e
        self.index = index
        e = self.op()
        return e

    def op(self):
        t = self.re(r"\w+")
        if t:
            c = self.condition()
            return NodePattern(t, c)
    
    def condition(self):
        # TODO: not implemented yet. This function is a placeholder
        index = self.index
        if self.token("["):
            c = self.token("1x1") or self.token("3x3")
            if c:
                if self.token("]"):
                    return c
            self.index = index
    
    def token(self, s):
        return self.re(r"\s*(" + re.escape(s) + r")\s*", 1)

    def string(self, s):
        if s == self.text[self.index:self.index+len(s)]:
            self.index += len(s)
            return s

    def re(self, regex, group=0):
        m = re.match(regex, self.text[self.index:])
        if m:
            self.index += len(m.group(0))
            return m.group(group)
            

class NodePattern():
    def __init__(self, op, condition=None):
        self.op = op
        self.condition = condition  # TODO: not implemented yet
    
    def match(self, graph, node):
        # if node is a list, means there exists multiple edges, pattern does not allow.
        if isinstance(node, list):
            return [], None
        if self.op == node.op and not node._skip_pattern_search:
            #  following nodes may be multiple, we only accept singleton
            following = graph.outgoing(node)
            if len(following) == 1:
                following = following[0]
            return [node], following
        else:
            return [], None


class SerialPattern():
    def __init__(self, patterns):
        self.patterns = patterns

    def match(self, graph, node):
        all_matches = []
        for i, p in enumerate(self.patterns):
            matches, following = p.match(graph, node)
            if not matches:
                return [], None
            all_matches.extend(matches)
            if i < len(self.patterns) - 1:
                node = following  # Might be more than one node
        return all_matches, following


class ParallelPattern():
    def __init__(self, patterns):
        self.patterns = patterns

    def match(self, graph, nodes):
        if not nodes:
            return [], None
        nodes = nodes if isinstance(nodes, list) else [nodes]
        # If a single node, assume we need to match with its siblings
        if len(nodes) == 1:
            nodes = graph.siblings(nodes[0])
        else:
            # Verify all nodes have the same parent or all have no parent
            parents = [graph.incoming(n) for n in nodes]
            matches = [set(p) == set(parents[0]) for p in parents[1:]]
            if not all(matches):
                return [], None

        # TODO: If more nodes than patterns, we should consider
        #       all permutations of the nodes
        if len(self.patterns) != len(nodes):
            return [], None
        
        patterns = self.patterns.copy()
        nodes = nodes.copy()
        all_matches = []
        end_node = None
        for p in patterns:
            found = False
            for n in nodes:
                matches, following = p.match(graph, n)
                if matches:
                    found = True
                    nodes.remove(n)
                    all_matches.extend(matches)
                    # Verify all branches end in the same node
                    if end_node:
                        if end_node != following:
                            return [], None
                    else:
                        end_node = following
                    break
            if not found:
                return [], None
        return all_matches, end_node





from setuptools import setup, find_packages

def parse_requirements(filename):
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

reqs = parse_requirements('requirements.txt')

VERSION = '2.0.11'
DESCRIPTION = 'Only Train Once (OTO): Automatic One-Shot General DNN Training and Compression Framework'
LONG_DESCRIPTION = 'Only Train Once (OTO): Automatic One-Shot General DNN Training and Compression (via Structured Pruning) Framework'

setup(
    name="only_train_once",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    author="Tianyi Chen",
    author_email="tiachen@microsoft.com",
    license='MIT',
    packages=find_packages(),
    install_requires=reqs,
    url="https://github.com/tianyic/only_train_once",
    keywords='automatic, one-shot, structure pruning, sparse optimization',
    classifiers= [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        'License :: OSI Approved :: MIT License',
        "Programming Language :: Python :: 3",
    ]
)


import os
import sys
import torch
import numpy as np
from torch.optim.optimizer import Optimizer, required
import torch.nn.functional as F
import time

currentdir = os.path.dirname(os.path.realpath(__file__))
parentdir = os.path.dirname(currentdir)
sys.path.append(parentdir)
from assets.group_types import GROUP_TYPE
from .hyperparameter import DEFAULT_OPT_PARAMS

class DHSPG(Optimizer):

    def __init__(self, params, variant='sgd', lr=required, lmbda=None, lmbda_amplify=None, hat_lmbda_coeff=None, epsilon=0.0, first_momentum=None, second_momentum=None, dampening=None, weight_decay=None, 
                 target_group_sparsity=0.5, tolerance_group_sparsity=0.05, start_pruning_steps=0, partition_step=None, half_space_project_steps=None, warm_up_steps=0, group_divisible=1, fixed_zero_groups=True):
        if lr is not required and lr < 0.0:
            raise ValueError("Invalid learning rate: {}".format(lr))

        if partition_step is None or half_space_project_steps is None:
            assert start_pruning_steps >= 0 
            self.partition_step = start_pruning_steps
            self.half_space_project_steps = start_pruning_steps
        else:
            self.partition_step = partition_step
            self.half_space_project_steps = half_space_project_steps
        
        # Set up hyper-parameters related to group sparsity exploration
        lmbda = lmbda if lmbda is not None else DEFAULT_OPT_PARAMS[variant]['lmbda']
        lmbda_amplify = lmbda_amplify if lmbda_amplify is not None else DEFAULT_OPT_PARAMS[variant]['lmbda_amplify']
        hat_lmbda_coeff = hat_lmbda_coeff if hat_lmbda_coeff is not None else DEFAULT_OPT_PARAMS[variant]['hat_lmbda_coeff']
        # Set up hyper-parameters related to baseline optimizer
        first_momentum = first_momentum if first_momentum is not None else DEFAULT_OPT_PARAMS[variant]['first_momentum']
        second_momentum = second_momentum if second_momentum is not None else DEFAULT_OPT_PARAMS[variant]['second_momentum']
        dampening = dampening if dampening is not None else DEFAULT_OPT_PARAMS[variant]['dampening']
        weight_decay = weight_decay if weight_decay is not None else DEFAULT_OPT_PARAMS[variant]['weight_decay']
        
        self.partitoned = False
        self.warm_up_steps = warm_up_steps
        
        self.fixed_zero_groups = fixed_zero_groups
        
        self.safe_guard = 1e-8
        self.target_group_sparsity = target_group_sparsity
        self.tolerance_group_sparsity = tolerance_group_sparsity
        self.x_norm_g_p_all = 0.0
        self.x_norm_g_np_all = 0.0
        self.num_groups_g_p_all = 0
        self.num_groups_g_np_all = 0

        defaults = dict(lr=lr, lmbda=lmbda, lmbda_amplify=lmbda_amplify, hat_lmbda_coeff=hat_lmbda_coeff, epsilon=epsilon,\
                        weight_decay=weight_decay, first_momentum=first_momentum, second_momentum=second_momentum, \
                        dampening=dampening, variant=variant, num_groups_g_p=0, num_groups_g_np=0, \
                        num_steps=0, partitoned=False, x_norm_g_p=0.0, x_norm_g_np=0.0, grad_variant=dict(),
                        global_start_idx=0, global_idx=0, group_divisible=group_divisible)
        super(DHSPG, self).__init__(params, defaults)

        self.auxilary_params = dict()
        for group in self.param_groups:
            if group['group_type'] == GROUP_TYPE['auxilary']:
                self.auxilary_params[group['cc_id']] = group

        self.curr_group_sparsity, _ = self.compute_group_sparsity_omega()

    def __setstate__(self, state):
        super(DHSPG, self).__setstate__(state)
   
    def update_xs(self, group, xs, new_flatten_xs):
        track_id = 0
        left_pointer = 0
        right_pointer = 0

        # tackle params belonging to the group itself
        for param in group['params']:
            if param.grad is None:
                continue
            right_pointer += xs[track_id].shape[1]      
            param.data.copy_(new_flatten_xs[:, left_pointer:right_pointer].contiguous().view(param.shape))
            track_id += 1
            left_pointer = right_pointer

        # tackle params belonging to auxiliary groups
        for cc_id, offset in group['auxiliary_ccs']:
            auxiliary_cc_group = self.auxilary_params[cc_id]
            for param in auxiliary_cc_group['params']:
                if param.grad is None:
                    continue
                right_pointer += xs[track_id].shape[1]
                param.data[offset:offset+group['num_groups'], ...] = new_flatten_xs[:, left_pointer:right_pointer].contiguous()\
                     .view(param[offset:offset+group['num_groups'], ...].shape)
                track_id += 1
                left_pointer = right_pointer

    def get_first_momentum_grad(self, param_state, key, momentum, dampening, grad):
        if momentum > 0:
            if key not in param_state:
                buf = param_state[key] = grad
            else:
                buf = param_state[key]
                buf.mul_(momentum).add_(grad, alpha=(1.0-dampening))
            return buf
        else:
            return grad 

    def get_second_momentum_grad_square(self, param_state, key, momentum, dampening, grad):
        if momentum > 0:
            if key not in param_state:
                buf = param_state[key] = grad * grad
            else:
                buf = param_state[key]
                buf.mul_(momentum).add_(grad * grad, alpha=(1.0-dampening))
            return buf
        else:
            return grad * grad

    def get_xs_grads(self, group, require_grad=True):
        xs = []
        grads = []
        for i, param in enumerate(group['params']):
            if param.grad is None and require_grad:
                continue
            if len(param.data.shape) == 1:
                xs.append(param.data.view(group['num_groups'], -1))
                if require_grad:
                    grads.append(group['grad_variant'][i].view(group['num_groups'], -1))
            elif len(param.data.shape) == 4: # conv layer
                xs.append(param.data.view(group['num_groups'], -1))
                if require_grad:
                    grads.append(group['grad_variant'][i].view(group['num_groups'], -1))
            else:
                xs.append(param.data.view(group['num_groups'], -1))
                if require_grad:
                    grads.append(group['grad_variant'][i].view(group['num_groups'], -1))

        for i, (cc_id, offset) in enumerate(group['auxiliary_ccs']):
            auxiliary_cc_group = self.auxilary_params[cc_id]
            for j, param in enumerate(auxiliary_cc_group['params']):
                if param.grad is None and require_grad:
                    continue
                auxilary_params = param.data[offset:offset + group['num_groups'], ...]
                if require_grad:
                    auxilary_grads = auxiliary_cc_group['grad_variant'][j][offset:offset + group['num_groups'], ...]
                if len(param.data.shape) == 1:
                    xs.append(auxilary_params.view(group['num_groups'], -1))
                    if require_grad:
                        grads.append(auxilary_grads.view(group['num_groups'], -1))
                elif len(param.data.shape) == 4: # conv layer
                    xs.append(auxilary_params.view(group['num_groups'], -1))
                    if require_grad:
                        grads.append(auxilary_grads.view(group['num_groups'], -1))
                else:
                    xs.append(auxilary_params.view(group['num_groups'], -1))
                    if require_grad:
                        grads.append(auxilary_grads.view(group['num_groups'], -1))
        return xs, grads

    def partition_groups(self):
        print("partition_groups")
        self.x_norm_g_p_all = 0.0
        self.x_norm_g_np_all = 0.0
        total_num_groups = 0
        global_cos_sims = []
        global_norm_groups_avg = []
        global_start_idx = 0
        for i, group in enumerate(self.param_groups):
            if group['group_type'] >= GROUP_TYPE['multi-head-linear']:
                xs, grads = self.get_xs_grads(group)
                flatten_x = torch.cat(xs, dim = 1)
                flatten_grad = torch.cat(grads, dim = 1)
                cos_sim = F.cosine_similarity(-flatten_x, -flatten_grad)
                group_size = flatten_x.shape[1] if len(flatten_x.shape) > 1 else 1
                norm_avg_groups = torch.norm(flatten_x, dim=1) / group_size
                global_cos_sims.append(cos_sim)
                global_norm_groups_avg.append(norm_avg_groups)
                total_num_groups += group['num_groups']
                group['global_start_idx'] = global_start_idx
                group['global_idx'] = np.arange(global_start_idx, global_start_idx+group['num_groups'])
                global_start_idx += group['num_groups']

        self.K = int(total_num_groups * min(self.target_group_sparsity + self.tolerance_group_sparsity, 0.999))
        global_cos_sims = torch.cat(global_cos_sims, dim=0)
        global_norm_groups_avg = torch.cat(global_norm_groups_avg, dim=0)

        cos_sims_scores = torch.nn.functional.normalize(global_cos_sims, dim=0)
        norms_scores = torch.nn.functional.normalize(1.0 / (global_norm_groups_avg + 1e-6), dim=0)
        scores = 0.5 * cos_sims_scores + 0.5 * norms_scores
        topk_scores, topk_indices = torch.topk(scores, self.K)

        topk_indices = topk_indices.cpu().numpy()
        
        self.num_groups_g_p_all = 0
        self.num_groups_g_np_all = 0
        
        for i, group in enumerate(self.param_groups):
            group['partitoned'] = True
            if group['group_type'] >= GROUP_TYPE['multi-head-linear']:
                xs, grads = self.get_xs_grads(group)
                flatten_x = torch.cat(xs, dim = 1)
                flatten_grad = torch.cat(grads, dim = 1)
                
                group['magnitude_penalize'] = np.intersect1d(topk_indices, group['global_idx'])
                group['not_magnitude_penalize'] = np.setdiff1d(group['global_idx'], group['magnitude_penalize'])

                group['magnitude_penalize'] -= group['global_start_idx']
                group['not_magnitude_penalize'] -= group['global_start_idx']

                if group['num_groups'] <= group['group_divisible']:
                    group['magnitude_penalize'] = []
                    group['not_magnitude_penalize'] = np.arange(0, group['num_groups'])
                else:
                    curr_num_np_groups = len(group['not_magnitude_penalize'])
                    ratio = float(curr_num_np_groups) / float(group['group_divisible'])
                    refined_num_np_groups = int(max(round(ratio), 1) * group['group_divisible'])
                    groups_scores = scores[group['global_start_idx']:group['global_start_idx']+group['num_groups']]
                    _, refined_np_groups_indices = torch.topk(-groups_scores, refined_num_np_groups)
                    refined_np_groups_indices = refined_np_groups_indices.cpu().numpy()
                    group['not_magnitude_penalize'] = refined_np_groups_indices
                    group['magnitude_penalize'] = np.setdiff1d(np.arange(0, group['num_groups']), group['not_magnitude_penalize'])            

                group['magnitude_penalize_bool'] = torch.zeros(group['num_groups'], dtype=torch.bool).cuda()
                group['magnitude_penalize_bool'][group['magnitude_penalize']] = True

                group['num_groups_g_p'] = len(group['magnitude_penalize'])
                group['num_groups_g_np'] = len(group['not_magnitude_penalize'])

                group['x_norm_g_p'] = torch.sum(torch.norm(flatten_x[group['magnitude_penalize']], p=2, dim=1))
                group['x_norm_g_np'] = torch.sum(torch.norm(flatten_x[group['not_magnitude_penalize']], p=2, dim=1))

                assert group['num_groups_g_p'] + group['num_groups_g_np'] == group['num_groups']

                self.x_norm_g_p_all += group['x_norm_g_p']
                self.x_norm_g_np_all += group['x_norm_g_np']
                self.num_groups_g_p_all += group['num_groups_g_p']
                self.num_groups_g_np_all += group['num_groups_g_np']
        
    def compute_grad_variant(self):
        for i, group in enumerate(self.param_groups):
            is_adam = group['variant'] == 'adam' or group['variant'] == 'adamw'
            first_bias_correction = 1.0 - group['first_momentum'] ** group['num_steps'] if is_adam else None
            second_bias_correction = 1.0 - group['second_momentum'] ** group['num_steps'] if is_adam else None
            
            group['grad_variant'] = dict()
            for j, p in enumerate(group['params']):
                if p.grad is None:
                    continue
                refined_grad_f = torch.clone(p.grad.data).detach()
                if group['weight_decay'] is not None and group['variant'] != 'adamw':
                    refined_grad_f += group['weight_decay'] * p.data
                if not is_adam:
                    if group['first_momentum'] > 0.0 or group['dampening'] > 0.0:
                        refined_grad_f = self.get_first_momentum_grad(group, f"grad_first_moment_buffer_group_{i}_param_{j}", 
                            group['first_momentum'], group['dampening'], refined_grad_f)
                    group['grad_variant'][j] = refined_grad_f
                else:
                    first_moment_grad = self.get_first_momentum_grad(group, f"grad_first_moment_buffer_group_{i}_param_{j}", 
                        group['first_momentum'], group['first_momentum'], refined_grad_f)                    
                    second_moment_grad_sq = self.get_second_momentum_grad_square(group, f"grad_second_moment_buffer_group_{i}_param_{j}", 
                        group['second_momentum'], group['second_momentum'], refined_grad_f)        

                    exp_avg_first_moment_grad = first_moment_grad / first_bias_correction
                    exp_avg_second_moment_grad_sq = second_moment_grad_sq / second_bias_correction
                    denom = exp_avg_second_moment_grad_sq.sqrt().add_(self.safe_guard)
                    group['grad_variant'][j] = exp_avg_first_moment_grad / denom

    def reach_target_group_sparsity(self):
        if self.curr_group_sparsity < self.target_group_sparsity:
            return False
        else:
            return True

    def step(self, closure=None):
        loss = None
        if closure is not None:
            loss = closure()
        
        for group in self.param_groups:
            group['num_steps'] += 1

        # First pass to compute gradient variant via different criteria
        self.compute_grad_variant()

        # Partition groups into G_p and G_np
        if not self.param_groups[0]['partitoned'] and self.param_groups[0]['num_steps'] - 1 == self.partition_step:
            print(self.param_groups[0]['num_steps'] - 1, self.partition_step)
            self.partition_groups()

        # Second pass to update variables
        for i, group in enumerate(self.param_groups):
            if group['group_type'] == GROUP_TYPE['no-update']:
                continue
            elif group['group_type'] == GROUP_TYPE['auxilary']:
                # The variables of auxiliary CC are updated within their stem CCs.
                continue
            elif group['group_type'] == GROUP_TYPE['default']:
                for j, p in enumerate(group['params']):
                    if p.grad is None:
                        continue
                    if group['weight_decay'] is not None and group['variant'] == 'adamw':
                        p.data.add_(group['weight_decay'] * p.data, alpha=-group['lr'])
                    p.data.add_(group['grad_variant'][j], alpha=-group['lr'])
            elif group['group_type'] == GROUP_TYPE['multi-head-linear']:
                # Since the transformer is facing some difficulty to convert onnx, 
                # we postpone the support of multi-head-attention temporally. 
                raise NotImplementedError
            elif group['group_type'] == GROUP_TYPE['linear'] or \
                 group['group_type'] == GROUP_TYPE['conv']:
                xs, grads = self.get_xs_grads(group)
                flatten_x = torch.cat(xs, dim = 1)
                flatten_grad = torch.cat(grads, dim = 1)
                flatten_hat_x = flatten_x - group['lr'] * flatten_grad

                if group['weight_decay'] is not None and group['variant'] == 'adamw':
                    flatten_hat_x = flatten_hat_x - group['lr'] * group['weight_decay'] * flatten_x
                
                flatten_x_norm = torch.norm(flatten_x, p=2, dim=1)
                # If some groups needs to penalize magnitude    
                if group['num_groups_g_p'] > 0 and not self.reach_target_group_sparsity():
                    # Use adjusted lambda to aggressively penalize magnitude 
                    if group['num_steps'] >= self.warm_up_steps:
                        flatten_grad_norm = torch.norm(flatten_grad, p=2, dim=1)            
                        flatten_x_grad_inner_prod = torch.sum(flatten_x * flatten_grad, dim=1)

                        lambdas = torch.ones_like(flatten_x_norm) * group['lmbda']
                        # If not reach target group sparsity, adjust lambda if needed
                        if not self.reach_target_group_sparsity():
                            # Groups need to adjust lambda 
                            groups_adjust_lambda = group['magnitude_penalize_bool'].cuda() & (flatten_x_grad_inner_prod < 0)
                            lambdas_lower_bound = -flatten_x_grad_inner_prod[groups_adjust_lambda] / flatten_x_norm[groups_adjust_lambda]
                            lambdas_upper_bound = -(flatten_grad_norm[groups_adjust_lambda] * flatten_grad_norm[groups_adjust_lambda] *\
                                                flatten_x_norm[groups_adjust_lambda] / flatten_x_grad_inner_prod[groups_adjust_lambda])
                            lambdas_adjust = torch.clip((group['lmbda_amplify'] * lambdas_lower_bound), \
                                                        min=group['lmbda'], \
                                                        max=group['lmbda'] * group['hat_lmbda_coeff']) 
                            exceeding_upper_bound = lambdas_adjust >= lambdas_upper_bound
                            lambdas_adjust[exceeding_upper_bound] = (lambdas_upper_bound[exceeding_upper_bound] + lambdas_lower_bound[exceeding_upper_bound]) / 2.0
                            lambdas[groups_adjust_lambda] = lambdas_adjust

                        grad_mixed_l1_l2 = torch.zeros_like(flatten_grad)
                        non_zero_idxes = flatten_x_norm > 0.0
                        grad_mixed_l1_l2[non_zero_idxes] = flatten_x[non_zero_idxes] / (flatten_x_norm[non_zero_idxes] + self.safe_guard).unsqueeze(1)

                        flatten_hat_x[group['magnitude_penalize_bool']] -= group['lr'] * lambdas[group['magnitude_penalize_bool']].unsqueeze(1) * grad_mixed_l1_l2[group['magnitude_penalize_bool']]

                        if group['num_steps'] >= self.half_space_project_steps:
                            flatten_hat_x[group['magnitude_penalize_bool']] = self.half_space_project(flatten_hat_x[group['magnitude_penalize_bool']], flatten_x[group['magnitude_penalize_bool']], group['epsilon'], group['upper_group_sparsity'])
                    # Use default lambda to moderately penalize magnitude  
                    else:
                        lambdas = torch.ones_like(flatten_x_norm) * group['lmbda']
                        grad_mixed_l1_l2 = torch.zeros_like(flatten_grad)
                        non_zero_idxes = flatten_x_norm > 0.0
                        grad_mixed_l1_l2[non_zero_idxes] = flatten_x[non_zero_idxes] / (flatten_x_norm[non_zero_idxes] + self.safe_guard).unsqueeze(1)
                        flatten_hat_x[group['magnitude_penalize_bool']] -= group['lr'] * lambdas[group['magnitude_penalize_bool']].unsqueeze(1) * grad_mixed_l1_l2[group['magnitude_penalize_bool']]                
                # Fixed non_free variables
                if group['num_steps'] > self.half_space_project_steps or self.fixed_zero_groups:                   
                    zero_group_idxes = flatten_x_norm == 0.0
                    flatten_hat_x[zero_group_idxes, ...] = 0.0
                self.update_xs(group, xs, flatten_hat_x)
            else:
                raise("some parameters are not in any group type, please check")
        
        self.curr_group_sparsity, _ = self.compute_group_sparsity_omega()
        return 

    def half_space_project(self, hat_x, x, epsilon, upper_group_sparsity):
        num_groups = x.shape[0]
        x_norm = torch.norm(x, p=2, dim=1)
        before_group_sparsity = torch.sum(x_norm == 0) / float(num_groups)
        if before_group_sparsity < upper_group_sparsity:
            proj_idx = (torch.bmm(hat_x.view(hat_x.shape[0], 1, -1), x.view(x.shape[0], -1, 1)).squeeze() \
                < epsilon * x_norm ** 2)    
            
            trial_group_sparsity = torch.sum(torch.logical_or(proj_idx, x_norm == 0)) / float(num_groups)
            # if trial group sparsity larger than upper group sparsity, then control the size of half-space projection
            if trial_group_sparsity > upper_group_sparsity:
                max_num_proj_groups = int(num_groups * (trial_group_sparsity - upper_group_sparsity))
                max_num_proj_groups = min(max(0, max_num_proj_groups), num_groups - 1)
                proj_group_idxes = torch.arange(num_groups)[proj_idx == True]
                refined_proj_idxes = torch.randperm(torch.sum(proj_idx))[:max_num_proj_groups].sort()[0]
                hat_x[proj_group_idxes[refined_proj_idxes], ...] = 0.0
            else:
                hat_x[proj_idx, ...] = 0.0
        return hat_x

    def compute_group_sparsity_omega(self):
        total_num_groups = torch.zeros(len(GROUP_TYPE)) + 1e-6
        total_num_zero_groups = torch.zeros(len(GROUP_TYPE))

        omega = torch.tensor(0.0)
        for group in self.param_groups:
            if group['group_type'] == GROUP_TYPE['no-update']:
                """Params that do not need update"""
                total_num_zero_groups[group['group_type']] = 0
            elif group['group_type'] == GROUP_TYPE['default']:
                """Params that are not included in the regularization"""
                total_num_zero_groups[group['group_type']] = 0
            elif group['group_type'] == GROUP_TYPE['auxilary']:
                pass
            elif group['group_type'] == GROUP_TYPE['multi-head-linear']:
                xs, _ = self.get_xs_grads(group, require_grad=False)
                flatten_x = torch.cat(xs, dim = 1)

                num_groups = flatten_x.shape[0] // group['num_heads']
                flatten_x = self.reorg_multi_head(flatten_x, num_groups, group['num_heads'])

                flatten_x_norm = torch.norm(flatten_x, p=2, dim=1)
                zero_group_idxes = flatten_x_norm == 0.0
                total_num_groups[group['group_type']] += num_groups
                total_num_zero_groups[group['group_type']] += torch.sum(zero_group_idxes).cpu()
                omega += torch.sum(flatten_x_norm)
            elif group['group_type'] == GROUP_TYPE['linear']:
                """Group for standard linear layer"""
                xs, _ = self.get_xs_grads(group, require_grad=False)
                flatten_x = torch.cat(xs, dim = 1)                       

                flatten_x_norm = torch.norm(flatten_x, p=2, dim=1)
                zero_group_idxes = flatten_x_norm == 0.0
                total_num_groups[group['group_type']] += flatten_x.shape[0]
                total_num_zero_groups[group['group_type']] += torch.sum(zero_group_idxes).cpu()
                omega += torch.sum(flatten_x_norm)
            elif group['group_type'] == GROUP_TYPE['conv']:
                xs, _ = self.get_xs_grads(group, require_grad=False)
                flatten_x = torch.cat(xs, dim = 1)                       

                flatten_x_norm = torch.norm(flatten_x, p=2, dim=1)
                zero_group_idxes = flatten_x_norm == 0.0
                total_num_groups[group['group_type']] += flatten_x.shape[0]
                total_num_zero_groups[group['group_type']] += torch.sum(zero_group_idxes).cpu()
                omega += torch.sum(flatten_x_norm)
        
        overall_group_sparsity = torch.sum(total_num_zero_groups) / torch.sum(total_num_groups)
        return overall_group_sparsity.cpu().numpy(), omega.cpu().numpy()

    def compute_norm_group_partitions(self):
        self.x_norm_g_p_all = 0.0
        self.x_norm_g_np_all = 0.0
        self.num_groups_g_p_all = 0
        self.num_groups_g_np_all = 0
        
        for i, group in enumerate(self.param_groups):
            if group['group_type'] >= GROUP_TYPE['multi-head-linear']:
                xs, _ = self.get_xs_grads(group, require_grad=False)
                flatten_x = torch.cat(xs, dim = 1)
                if 'magnitude_penalize' not in group or 'not_magnitude_penalize' not in group:
                    return 0.0, 0.0, 0, 0
                group['x_norm_g_p'] = torch.sum(torch.norm(flatten_x[group['magnitude_penalize']], p=2, dim=1))
                group['x_norm_g_np'] = torch.sum(torch.norm(flatten_x[group['not_magnitude_penalize']], p=2, dim=1))
                self.x_norm_g_p_all += group['x_norm_g_p']
                self.x_norm_g_np_all += group['x_norm_g_np']
                self.num_groups_g_p_all += group['num_groups_g_p']
                self.num_groups_g_np_all += group['num_groups_g_np']
        return self.x_norm_g_p_all.cpu().item(), self.x_norm_g_np_all.cpu().item(), self.num_groups_g_p_all, self.num_groups_g_np_all

    def set_learning_rate(self, lr):
        for param_group in self.param_groups:
            param_group['lr'] = lr

    def get_learning_rate(self):
        for param_group in self.param_groups:
            lr = param_group['lr']
        return lr

    def set_lmbda(self, lmbda):
        for param_group in self.param_groups:
            param_group['lmbda'] = lmbda

    def get_lmbda(self):
        lmbda = 0.0
        for param_group in self.param_groups:
            lmbda = param_group['lmbda']
        return lmbda

    def set_hat_lmbda_coeff(self, hat_lmbda_coeff):
        for param_group in self.param_groups:
            param_group['hat_lmbda_coeff'] = hat_lmbda_coeff

    def get_hat_lmbda_coeff(self):
        hat_lmbda_coeff = 0.0
        for param_group in self.param_groups:
            hat_lmbda_coeff = param_group['hat_lmbda_coeff']
        return hat_lmbda_coeff

    def set_lmbda_amplify(self, lmbda_amplify):
        for param_group in self.param_groups:
            param_group['lmbda_amplify'] = lmbda_amplify

    def get_lmbda_amplify(self):
        lmbda_amplify = 0.0
        for param_group in self.param_groups:
            lmbda_amplify = param_group['lmbda_amplify']
        return lmbda_amplify

    def set_eps(self, eps):
        for param_group in self.param_groups:
            param_group['epsilon'] = eps
    
    def get_eps(self):
        eps = -1
        for param_group in self.param_groups:
            eps = max(eps, param_group['epsilon'])
        return eps



import imp
from .graph import Graph
from .zig.zig import automated_partition_zigs
from .optimizer import DHSPG
from .compression.compression import automated_compression
import os
from .flops.flops import compute_flops

class OTO:
    def __init__(self, model=None, dummy_input=None):
        self._graph = None
        self._model = model
        self._dummy_input = dummy_input

        if self._model is not None and self._dummy_input is not None:
            self.initialize(model=self._model, dummy_input=self._dummy_input)
            self.partition_zigs()
        self.compressed_model_path = None
        self.full_group_sparse_model_path = None

    def initialize(self, model=None, dummy_input=None):
        model = model.eval()
        self._model = model
        self._dummy_input = dummy_input
        self._graph = Graph(model, dummy_input)

    def partition_zigs(self):
        self._graph = automated_partition_zigs(self._graph)
    
    def visualize_zigs(self, out_dir=None, view=False, vertical=True):
        self._graph.build_dot(verbose=True, vertical=vertical).render(\
            os.path.join(out_dir if out_dir is not None else './', \
                self._model.name if hasattr(self._model, 'name') else type(self._model).__name__ + '_zig.gv'), \
                view=view)

    def dhspg(self, lr=0.1, lmbda=None, lmbda_amplify=None, hat_lmbda_coeff=None, epsilon=0.0, weight_decay=None, first_momentum=None, second_momentum=None, \
               variant='sgd', target_group_sparsity=0.5, tolerance_group_sparsity=0.05, partition_step=None, start_pruning_steps=0, half_space_project_steps=None,\
               warm_up_steps=0, dampening=None, group_divisible=1, fixed_zero_groups=True):
        self._optimizer = DHSPG(
            params=self._graph.params_groups(epsilon=epsilon),
            lr=lr,
            lmbda=lmbda,
            lmbda_amplify=lmbda_amplify,
            hat_lmbda_coeff=hat_lmbda_coeff,
            epsilon=epsilon,
            weight_decay=weight_decay,
            first_momentum=first_momentum,
            second_momentum=second_momentum,
            dampening=dampening,
            variant=variant,
            target_group_sparsity=target_group_sparsity, 
            tolerance_group_sparsity=tolerance_group_sparsity,
            partition_step=partition_step,
            warm_up_steps=warm_up_steps,
            start_pruning_steps=start_pruning_steps,
            half_space_project_steps=half_space_project_steps,
            group_divisible=group_divisible,
            fixed_zero_groups=fixed_zero_groups)
        return self._optimizer

    def compress(self, compressed_model_path=None, dummy_input=None, dynamic_axes=[False, dict()]):
        _, self.compressed_model_path, self.full_model_path = automated_compression(
            oto_graph=self._graph,
            model=self._model,
            dummy_input=self._dummy_input if dummy_input is None else dummy_input,
            compressed_model_path=compressed_model_path,
            dynamic_axes=dynamic_axes)
    
    def random_set_zero_groups(self):
        self._graph.random_set_zero_groups()
    
    def compute_flops(self, compressed=False, verbose=False):
        flops_info = compute_flops(self._graph, compressed=compressed)
        return flops_info['total'] if not verbose else flops_info
        
    def compute_num_params(self, compressed=False):
        import onnx
        import numpy as np
        if compressed:
            if self.compressed_model_path is None:
                raise "Compressed model does not exist, please compress at first."            
            onnx_model = onnx.load(self.compressed_model_path)
            onnx_graph = onnx_model.graph
            num_params = 0
            for tensor in onnx_graph.initializer:
                num_params += np.prod(tensor.dims)
            return num_params
        else:
            return sum([w.numel() for _, w in self._model.named_parameters()])